# coding=utf-8
from __future__ import absolute_import, print_function

import itertools

from suanpan.arguments import Int, String
from suanpan.dw import hive, mysql, odps, postgres
from suanpan.proxy import Proxy


class DataWarehouseProxy(Proxy):
    MAPPING = {
        "hive": hive.DataWarehouse,
        "odps": odps.DataWarehouse,
        "postgres": postgres.DataWarehouse,
        "mysql": mysql.DataWarehouse,
    }
    DEFAULT_ARGUMENTS = [String("dw-type")]
    HIVE_ARGUMENTS = [
        String("dw-hive-host", default="localhost"),
        Int("dw-hive-port", default=10000),
        String("dw-hive-database", default="default"),
        String("dw-hive-username"),
        String("dw-hive-password"),
        String("dw-hive-auth"),
    ]

    ODPS_ARGUMENTS = [
        String("dw-odps-access-id"),
        String("dw-odps-access-key"),
        String(
            "dw-odps-endpoint", default="http://service.cn.maxcompute.aliyun.com/api"
        ),
        String("dw-odps-project"),
    ]

    POSTGRES_ARGUMENTS = [
        String("dw-postgres-host", default="localhost"),
        Int("dw-postgres-port", default=5432),
        String("dw-postgres-database"),
        String("dw-postgres-username"),
        String("dw-postgres-password"),
    ]

    MYSQL_ARGUMENTS = [
        String("dw-mysql-host", default="localhost"),
        Int("dw-mysql-port", default=3306),
        String("dw-mysql-database"),
        String("dw-mysql-username"),
        String("dw-mysql-password"),
    ]

    ARGUMENTS = list(
        itertools.chain(
            DEFAULT_ARGUMENTS,
            HIVE_ARGUMENTS,
            ODPS_ARGUMENTS,
            POSTGRES_ARGUMENTS,
            MYSQL_ARGUMENTS,
        )
    )


dw = DataWarehouseProxy()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import api


def listComponents(limit=9999):
    return api.affinity.post("/component/list", json={"limit": limit})["list"]


def shareComponent(componentId, userId, name):
    return api.affinity.post("/component/share", json={"id": componentId, "targetUserId": userId, "name": name})



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import api


def getAccessKey():
    return api.affinity.get("/oss/ak")


def getToken():
    return api.affinity.get("/oss/token")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import String
from suanpan.dw import dw
from suanpan.tools import ToolComponent as tc
from suanpan.utils import csv


@tc.param(String(key="file", required=True))
@tc.param(String(key="table", required=True))
def SPDWUpload(context):
    args = context.args
    dw.writeTable(args.table, csv.load(args.file))


if __name__ == "__main__":
    SPDWUpload()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import String
from suanpan.storage import storage
from suanpan.tools import ToolComponent as tc


@tc.param(String(key="src", required=True))
@tc.param(String(key="dist", required=True))
def SPStorageCopy(context):
    args = context.args
    storage.copy(args.src, args.dist)


if __name__ == "__main__":
    SPStorageCopy()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, division, print_function

import os
import tempfile

from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base


class Storage(base.Storage):
    def __init__(self, localTempStore=tempfile.gettempdir(), **kwargs):
        super(Storage, self).__init__(
            delimiter=os.sep, tempStore=localTempStore, **kwargs
        )

    def download(self, name, path, workers=None):  # pylint: disable=unused-argument
        logger.info(
            f"Downloading: {self.storageUrl(name)} -> {path} - ({self.name}) Nothing to do!"
        )
        return self.getPathInTempStore(path)

    def upload(self, name, path, workers=None):  # pylint: disable=unused-argument
        logger.info(
            f"Uploading: {self.storageUrl(name)} -> {path} - ({self.name}) Nothing to do!"
        )
        return path

    def copy(self, path, dist, workers=None):  # pylint: disable=unused-argument
        pathUrl = self.storageUrl(path)
        distUrl = self.storageUrl(dist)
        logger.info(f"Copying: {pathUrl} -> {distUrl}")
        _path = self.getPathInTempStore(path)
        _dist = self.getPathInTempStore(dist)
        spath.copy(_path, _dist)
        logger.info(f"Copied: {pathUrl} -> {distUrl}")
        return dist

    def remove(self, path, workers=None):  # pylint: disable=unused-argument
        spath.remove(self.getPathInTempStore(path))
        logger.info(f"Removed: {self.storageUrl(path)}")
        return path

    def walk(self, folderName):
        root = self.getPathInTempStore(folderName)
        return runtime.saferun(os.walk, default=iter(()))(root)

    def listAll(self, folderName):
        root = self.getPathInTempStore(folderName)
        return (self.storagePathJoin(root, p) for p in os.listdir(root))

    def listFolders(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isdir(p))

    def listFiles(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isfile(p))

    def isFolder(self, folderName):
        folder = self.getPathInTempStore(folderName)
        return os.path.isdir(folder)

    def isFile(self, objectName):
        file = self.getPathInTempStore(objectName)
        return os.path.isfile(file)

    def storagePathJoin(self, *paths):
        return self.localPathJoin(*paths)

    def storageRelativePath(self, path, basepath):
        return self.localRelativePath(path, basepath)

    def storageUrl(self, path):
        return "file://" + self.getPathInTempStore(path)

    def getStorageMd5(self, objectName):
        return self.getLocalMd5(self.getPathInTempStore(objectName))

    def getStorageSize(self, objectName):
        return self.getLocalSize(self.getPathInTempStore(objectName))



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.api import app, component, oss
from suanpan.api.requests import affinity, host

get = host.get
post = host.post
put = host.put
delete = host.delete
call = host.call



# coding=utf-8
from __future__ import absolute_import, print_function

import contextlib
import itertools
from collections import defaultdict

from suanpan import error, runtime
from suanpan.arguments.auto import AutoArg
from suanpan.interfaces import (HasArguments, HasCallHooks, HasExitHooks,
                                HasInitHooks, HasSaveHooks)
from suanpan.log import logger
from suanpan.objects import Context


class Arguments(Context):
    pass


class Result(Context):
    pass


class Component(HasArguments, HasInitHooks, HasSaveHooks, HasCallHooks, HasExitHooks):
    def __init__(self, funcOrComponent):
        if isinstance(funcOrComponent, Component):
            self.runFunc = funcOrComponent.runFunc
            self.arguments = funcOrComponent.arguments
            self.beforeInitHooks = funcOrComponent.beforeInitHooks
            self.afterInitHooks = funcOrComponent.afterInitHooks
            self.beforeCallHooks = funcOrComponent.beforeCallHooks
            self.afterCallHooks = funcOrComponent.afterCallHooks
            self.beforeSaveHooks = funcOrComponent.beforeSaveHooks
            self.afterSaveHooks = funcOrComponent.afterSaveHooks
        else:
            super(Component, self).__init__()
            self.runFunc = funcOrComponent
            self.arguments = defaultdict(list)

    def __call__(self, *arg, **kwargs):
        self.run(*arg, **kwargs)
        return self

    def start(self):
        pass

    @property
    def name(self):
        return self.runFunc.__name__

    @runtime.globalrun
    def run(self, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(*arg, **kwargs)
        self.callAfterInitHooks(context)
        self.registerBeforeExitHooks(context)

        self.callBeforeCallHooks(context)
        results = self.runFunc(context)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(context, results)
        self.callAfterSaveHooks(context)

        return outputs

    def init(self, *arg, **kwargs):
        restArgs = self.getArgList(*arg, **kwargs)
        globalArgs, restArgs = self.loadGlobalArguments(restArgs=restArgs)
        self.initBase(globalArgs)
        context = self._getContext(globalArgs)
        args, restArgs = self.loadComponentArguments(context, restArgs=restArgs)
        context.update(args=args)
        return context

    def initBase(self, args):
        pass

    def save(self, context, results):
        outputs = self.saveOutputs(context, results)
        self._closeContext()
        return outputs

    @contextlib.contextmanager
    def context(self, args=None):  # pylint: disable=unused-argument
        yield Context()

    def _getContext(self, *args, **kwargs):
        self.contextManager = self.context(  # pylint: disable=attribute-defined-outside-init
            *args, **kwargs
        )
        return next(self.contextManager.gen)  # pylint: disable-msg=e1101

    def _closeContext(self):
        try:
            next(self.contextManager.gen)  # pylint: disable-msg=e1101
        except StopIteration:
            pass

    def loadGlobalArguments(self, restArgs=None, **kwargs):
        logger.info("Loading Global Arguments:")
        arguments, restArgs = super(Component, self).loadGlobalArguments(
            restArgs=restArgs, **kwargs
        )
        args = self.argumentsDict(arguments)
        return Arguments.froms(args), restArgs

    def loadComponentArguments(self, context, restArgs=None):
        logger.info("Loading Component Arguments:")
        _, restArgs = self.loadFormatArguments(
            context, restArgs=restArgs, exclude="outputs"
        )
        _, restArgs = self.loadCleanArguments(
            context, restArgs=restArgs, include="outputs"
        )
        arguments = {
            k: self.argumentsDict(arg for arg in v if arg.isSet)
            for k, v in self.arguments.items()
        }
        arguments.update(self.argumentsDict(itertools.chain(*self.arguments.values())))
        return Arguments.froms(arguments), restArgs

    def getArguments(self, include=None, exclude=None):
        includes = set(self.arguments.keys() if not include else self._list(include))
        excludes = set([] if not exclude else self._list(exclude))
        includes = includes - excludes
        argumentsLists = [self.arguments[c] for c in includes]
        return list(itertools.chain(*argumentsLists))

    def saveMutipleOutputs(self, context, outputs, results):
        if isinstance(results, (tuple, list)):
            results = (Result.froms(value=result) for result in results)
            return self.saveArguments(context, outputs, results)
        if isinstance(results, dict):
            outputs, results = zip(
                *[
                    (
                        argument,
                        Result.froms(
                            value=self.getArgumentValueFromDict(results, argument)
                        ),
                    )
                    for argument in outputs
                    if self.getArgumentValueFromDict(results, argument) is not None
                ]
            )
            return self.saveArguments(context, outputs, results)
        raise error.ComponentError(f"Incorrect results {results}")

    def saveOneOutput(self, context, output, results):
        result = (
            Result.froms(value=self.getArgumentValueFromDict(results, output))
            if isinstance(results, dict)
            and self.hasArgumentValueFromDict(results, output)
            else Result.froms(value=results)
        )
        return {output.key: output.save(context, result)}

    def saveOutputs(self, context, results):
        logger.info("Saving...")
        outputs = self.getArguments(include="outputs")
        if len(outputs) > 1:
            return self.saveMutipleOutputs(context, outputs, results)
        if len(outputs) == 1:
            return self.saveOneOutput(context, outputs[0], results)
        return None

    def addArgument(self, arg, argtype="args", reverse=True):
        if reverse:
            self.arguments[argtype].insert(0, arg)
        else:
            self.arguments[argtype].append(arg)

    @classmethod
    def use(cls, funcOrComponent):
        return (
            funcOrComponent
            if isinstance(funcOrComponent, cls)
            else cls(funcOrComponent)
        )

    @classmethod
    def arg(cls, argument, argtype="args", reverse=True):
        def _dec(funcOrComponent):
            funcOrComponent = (
                funcOrComponent
                if isinstance(funcOrComponent, cls)
                else cls(funcOrComponent)
            )
            if isinstance(argument, AutoArg):
                argument.setBackend(argtype=argtype)
            funcOrComponent.addArgument(argument, argtype=argtype, reverse=reverse)
            return funcOrComponent

        return _dec

    @classmethod
    def input(cls, *args, **kwargs):
        kwargs.update(argtype="inputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def output(cls, *args, **kwargs):
        kwargs.update(argtype="outputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def param(cls, *args, **kwargs):
        kwargs.update(argtype="params")
        return cls.arg(*args, **kwargs)

    @classmethod
    def column(cls, *args, **kwargs):
        kwargs.update(argtype="columns")
        return cls.arg(*args, **kwargs)

    def _list(self, params=None):
        return [params] if isinstance(params, str) else list(params)



# coding=utf-8
from __future__ import absolute_import, print_function

import argparse
import atexit
import base64
import functools
import itertools
import re
import signal
import sys

from suanpan import g
from suanpan.arguments import Bool
from suanpan.log import logger
from suanpan.objects import Context, HasName


class HasArguments(object):
    GLOBAL_ARGUMENTS = []
    ARGUMENTS = []

    @classmethod
    def dictFromList(cls, l):
        return dict(zip(l[0::2], l[1::2]))

    @classmethod
    def listFromDict(cls, d):
        return list(itertools.chain(*d.items()))

    @classmethod
    def getArgList(cls, *args, **kwargs):  # pylint: disable=unused-argument
        argDict = {}
        argDict.update(cls.dictFromList(cls.getArgListFromEnv()))
        argDict.update(cls.dictFromList(cls.getArgListFromCli()))
        argDict.update(cls.dictFromList(args))
        argDict.update({f"--{k}": v for k, v in kwargs.items()})
        return cls.listFromDict(argDict)

    @classmethod
    def getArgListFromCli(cls):
        return sys.argv[1:]

    @classmethod
    def getArgListFromEnv(cls):
        argStringBase64 = g.appParams
        logger.debug(f"SP_PARAM(Base64)='{argStringBase64}'")
        try:
            argString = base64.b64decode(argStringBase64).decode()
        except Exception:  # pylint: disable=broad-except
            argString = argStringBase64  # temporary fix for SP_PARAM(Base64)
        logger.debug(f"SP_PARAM='{argString}'")
        regex = r"(--[\w-]+)\s+(?:(?P<quote>[\"\'])(.*?)(?P=quote)|([^\'\"\s]+))"
        groups = re.findall(regex, argString, flags=re.S)
        argv = list(
            itertools.chain(*[(group[0], group[-2] or group[-1]) for group in groups])
        )
        return argv

    def resetArguments(self, *args, **kwargs):
        arguments = self.getArguments(*args, **kwargs)
        return self._resetArguments(arguments)

    def loadFormatArguments(self, context, restArgs=None, **kwargs):
        arguments = self.getArguments(**kwargs)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(context, arguments)
        return arguments, restArgs

    def loadCleanArguments(self, context, restArgs=None, **kwargs):
        arguments = self.getArguments(**kwargs)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._cleanArguments(context, arguments)
        return arguments, restArgs

    def loadGlobalArguments(self, restArgs=None, **kwargs):
        arguments = self.getGlobalArguments(**kwargs)
        args, restArgs = self._parseGlobalArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(Context(), arguments)
        return arguments, restArgs

    def loadComponentArguments(self, restArgs=None, **kwargs):
        arguments = self.getComponentArguments(**kwargs)
        self._resetArguments(arguments)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(Context(), arguments)
        return arguments, restArgs

    def saveArguments(self, context, arguments, results):
        return {
            argument.key: argument.save(context, result)
            for argument, result in zip(arguments, results)
        }

    def getArguments(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    def getGlobalArguments(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.GLOBAL_ARGUMENTS

    def getComponentArguments(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.ARGUMENTS

    @classmethod
    def _parseArguments(cls, arguments, restArgs=None, **kwargs):
        parser = argparse.ArgumentParser(allow_abbrev=False, **kwargs)
        for arg in arguments:
            arg.addParserArguments(parser)
        return parser.parse_known_args(restArgs)

    @classmethod
    def _parseGlobalArguments(cls, arguments, restArgs=None, **kwargs):
        parser = argparse.ArgumentParser(**kwargs)
        for arg in arguments:
            arg.addGlobalParserArguments(parser)
        return parser.parse_known_args(restArgs)

    @classmethod
    def _resetArguments(cls, arguments):
        return {arg.key: arg.reset() for arg in arguments}

    @classmethod
    def _loadArguments(cls, args, arguments):
        return {arg.key: arg.load(args) for arg in arguments}

    @classmethod
    def _formatArguments(cls, context, arguments):
        return {arg.key: arg.format(context) for arg in arguments}

    @classmethod
    def _cleanArguments(cls, context, arguments):
        return {arg.key: arg.clean(context) for arg in arguments}

    @classmethod
    def defaultArgumentsFormat(cls, args, arguments):
        arguments = (arg.key.replace("-", "_") for arg in arguments)
        return {
            cls._defaultArgumentKeyFormat(arg): getattr(args, arg, None)
            for arg in arguments
        }

    @classmethod
    def _defaultArgumentKeyFormat(cls, key):
        return cls._toCamelCase(cls._removePrefix(key))

    @classmethod
    def _removePrefix(cls, string, delimiter="_", num=1):
        pieces = string.split(delimiter)
        pieces = pieces[num:] if len(pieces) > num else pieces
        return delimiter.join(pieces)

    @classmethod
    def _toCamelCase(cls, string, delimiter="_"):
        camelCaseUpper = lambda i, s: s[0].upper() + s[1:] if i and s else s
        return "".join(
            [camelCaseUpper(i, s) for i, s in enumerate(string.split(delimiter))]
        )

    @classmethod
    def argumentsDict(cls, arguments):
        result = {}
        for arg in arguments:
            keys = (arg.key, arg.alias)
            result.update({key: arg.value for key in keys if key})
        return result

    @classmethod
    def getArgumentValueFromDict(cls, data, arg):
        value = data.get(arg.alias)
        if value is not None:
            return value

        value = data.get(arg.key)
        if value is not None:
            return value

        return None

    @classmethod
    def hasArgumentValueFromDict(cls, data, arg):
        return arg.alias in data or arg.key in data


class HasLogger(HasName):
    def __init__(self):
        super(HasLogger, self).__init__()
        logger.setLogger(self.name)

    @property
    def logger(self):
        return logger


class HasDevMode(HasArguments):
    DEV_ARGUMENTS = [Bool(key="debug", default=False)]

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(HasDevMode, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.DEV_ARGUMENTS


class HasInitHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasInitHooks, self).__init__()
        self.beforeInitHooks = getattr(self, "beforeInitHooks", [])
        self.afterInitHooks = getattr(self, "afterInitHooks", [])

    def beforeInit(self, *args, **kwargs):
        pass

    def callBeforeInitHooks(self, *args, **kwargs):
        self.beforeInit(*args, **kwargs)
        for hook in self.beforeInitHooks:
            hook(*args, **kwargs)

    def addBeforeInitHooks(self, *hooks):
        self.beforeInitHooks.extend(hooks)
        return self

    def afterInit(self, *args, **kwargs):
        pass

    def callAfterInitHooks(self, *args, **kwargs):
        self.afterInit(*args, **kwargs)
        for hook in self.afterInitHooks:
            hook(*args, **kwargs)

    def addAfterInitHooks(self, *hooks):
        self.afterInitHooks.extend(hooks)
        return self


class HasSaveHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasSaveHooks, self).__init__()
        self.beforeSaveHooks = getattr(self, "beforeSaveHooks", [])
        self.afterSaveHooks = getattr(self, "afterSaveHooks", [])

    def beforeSave(self, *args, **kwargs):
        pass

    def callBeforeSaveHooks(self, *args, **kwargs):
        self.beforeSave(*args, **kwargs)
        for hook in self.beforeSaveHooks:
            hook(*args, **kwargs)

    def addBeforeSaveHooks(self, *hooks):
        self.beforeSaveHooks.extend(hooks)
        return self

    def afterSave(self, *args, **kwargs):
        pass

    def callAfterSaveHooks(self, *args, **kwargs):
        self.afterSave(*args, **kwargs)
        for hook in self.afterSaveHooks:
            hook(*args, **kwargs)

    def addAfterSaveHooks(self, *hooks):
        self.afterSaveHooks.extend(hooks)
        return self


class HasCallHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasCallHooks, self).__init__()
        self.beforeCallHooks = getattr(self, "beforeCallHooks", [])
        self.afterCallHooks = getattr(self, "afterCallHooks", [])

    def beforeCall(self, *args, **kwargs):
        pass

    def callBeforeCallHooks(self, *args, **kwargs):
        self.beforeCall(*args, **kwargs)
        for hook in self.beforeCallHooks:
            hook(*args, **kwargs)

    def addBeforeCallHooks(self, *hooks):
        self.beforeCallHooks.extend(hooks)
        return self

    def afterCall(self, *args, **kwargs):
        pass

    def callAfterCallHooks(self, *args, **kwargs):
        self.afterCall(*args, **kwargs)
        for hook in self.afterCallHooks:
            hook(*args, **kwargs)

    def addAfterCallHooks(self, *hooks):
        self.afterCallHooks.extend(hooks)
        return self


class HasTriggerHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasTriggerHooks, self).__init__()
        self.beforeTriggerHooks = getattr(self, "beforeTriggerHooks", [])
        self.afterTriggerHooks = getattr(self, "afterTriggerHooks", [])

    def beforeTrigger(self, *args, **kwargs):
        pass

    def callBeforeTriggerHooks(self, *args, **kwargs):
        self.beforeTrigger(*args, **kwargs)
        for hook in self.beforeTriggerHooks:
            hook(*args, **kwargs)

    def addBeforeTriggerHooks(self, *hooks):
        self.beforeTriggerHooks.extend(hooks)
        return self

    def afterTrigger(self, *args, **kwargs):
        pass

    def callAfterTriggerHooks(self, *args, **kwargs):
        self.afterTrigger(*args, **kwargs)
        for hook in self.afterTriggerHooks:
            hook(*args, **kwargs)

    def addAfterTriggerHooks(self, *hooks):
        self.afterTriggerHooks.extend(hooks)
        return self


class HasExitHooks(object):
    SIGNALS = (
        signal.SIGUSR1,
        signal.SIGTERM,
        signal.SIGQUIT,
        signal.SIGHUP,
        signal.SIGINT,
    )

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasExitHooks, self).__init__()
        self.beforeExitHooks = getattr(self, "beforeExitHooks", [])

    def _exit(self, signum, frame):
        logger.debug(f"Signal Exit: {signum} {frame}")
        sys.exit(1)

    def beforeExit(self, *args, **kwargs):
        pass

    def callBeforeExitHooks(self, *args, **kwargs):
        self.beforeExit(*args, **kwargs)
        for hook in self.beforeExitHooks:
            hook(*args, **kwargs)

    def addBeforeExitHooks(self, *hooks):
        self.beforeExitHooks.extend(hooks)
        return self

    def registerBeforeExitHooks(self, *args, **kwargs):
        atexit.register(self.callBeforeExitHooks, *args, **kwargs)
        for s in self.SIGNALS:
            signal.signal(s, self._exit)
        return self



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.interfaces import HasCallHooks, HasExitHooks, HasInitHooks
from suanpan.objects import HasName


class BaseApp(HasName, HasInitHooks, HasCallHooks, HasExitHooks):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    def start(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    @property
    def trigger(self):
        raise NotImplementedError(f"{self.name} not support trigger")

    def input(self, argument):
        raise NotImplementedError("Method not implemented!")

    def output(self, argument):
        raise NotImplementedError("Method not implemented!")

    def param(self, argument):
        raise NotImplementedError("Method not implemented!")

    def column(self, argument):
        raise NotImplementedError("Method not implemented!")

    def beforeInit(self, hook):
        self.addBeforeInitHooks(hook)
        return hook

    def afterInit(self, hook):
        self.addAfterInitHooks(hook)
        return hook

    def beforeCall(self, hook):
        self.addBeforeCallHooks(hook)
        return hook

    def afterCall(self, hook):
        self.addAfterCallHooks(hook)
        return hook

    def beforeExit(self, hook):
        self.addBeforeExitHooks(hook)
        return self

    def send(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support send")

    @property
    def args(self):
        raise NotImplementedError(f"{self.name} not support args")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.proxy import Proxy


class ProxyArg(Proxy):
    TYPE_KEY = "argtype"


class AutoArg(ProxyArg):
    DEFAULT_TYPE = "params"



# coding=utf-8
from __future__ import absolute_import, print_function

import logging
import logging.handlers
import os
import platform
import time

import suanpan
from suanpan import path, g
from suanpan.utils import functional


class Formatter(logging.Formatter):

    converter = time.gmtime

    def __init__(
        self,
        fmt="%(asctime)s.%(msecs)06d :: %(levelname)-10s :: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    ):
        super(Formatter, self).__init__(fmt=fmt, datefmt=datefmt)


class Logger(logging.Logger):

    FORMATTER = Formatter()
    STREAM_LOG_LEVEL = logging.DEBUG
    FILE_LOG_LEVEL = logging.DEBUG
    LOG_FILE_MAX_SIZE = 1024 * 1024 * 1024
    BACKUP_COUNT = 5
    LOG_PATH = "logs"
    LOG_FILE = None

    def __init__(self, name="suanpan"):
        super(Logger, self).__init__(name=name)
        self.addStreamHandler(level=logging.DEBUG if g.debug else logging.INFO)
        # self.addFileHandler()

    def addStreamHandler(self, level=STREAM_LOG_LEVEL, formatter=FORMATTER):
        streamHandler = logging.StreamHandler()
        streamHandler.setLevel(level)
        streamHandler.setFormatter(formatter)
        self.addHandler(streamHandler)
        return streamHandler

    def addFileHandler(
        self,
        level=FILE_LOG_LEVEL,
        formatter=FORMATTER,
        logPath=LOG_PATH,
        logFile=LOG_FILE,
        logFileMaxSize=LOG_FILE_MAX_SIZE,
        backupCount=BACKUP_COUNT,
    ):
        logFile = logFile or f"{self.name}.log"
        logFilePath = os.path.join(logPath, logFile)
        path.mkdirs(logFilePath, parent=True)
        fileHandler = logging.handlers.RotatingFileHandler(
            logFilePath, maxBytes=logFileMaxSize, backupCount=backupCount
        )
        fileHandler.setLevel(level)
        fileHandler.setFormatter(formatter)
        self.addHandler(fileHandler)
        return fileHandler

    @functional.onlyonce
    def logDebugInfo(self):
        self.logPythonVersion()
        self.logSdkVersion()

    @functional.onlyonce
    def logSdkVersion(self):
        self.debug(f"Suanpan SDK (ver: {suanpan.__version__})")

    @functional.onlyonce
    def logPythonVersion(self):
        self.debug(f"Python (ver: {platform.python_version()})")


class LoggerProxy(object):
    def __init__(self, loggerOrName):
        self.setLogger(loggerOrName)

    def __getattr__(self, key):
        return getattr(self.logger, key)

    def setLogger(self, loggerOrName):
        self.logger = (
            loggerOrName if isinstance(loggerOrName, Logger) else Logger(loggerOrName)
        )


rootLogger = Logger("suanpan")
logger = LoggerProxy(rootLogger)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.docker import DockerComponent
from suanpan.log import logger


class ToolComponent(DockerComponent):
    ENABLED_BASE_SERVICES = {"dw", "storage"}

    def beforeInit(self):
        logger.logDebugInfo()
        logger.debug(f"ToolComponent {self.name} starting...")

    def afterSave(self, context):  # pylint: disable=unused-argument
        logger.debug(f"ToolComponent {self.name} done.")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error
from suanpan.app.base import BaseApp
from suanpan.spark import SparkComponent as Handler


class SparkApp(BaseApp):
    def __init__(self, *args, **kwargs):
        super(SparkApp, self).__init__(*args, **kwargs)
        self.handler = None

    def __call__(self, *args, **kwargs):
        if not self.handler:
            raise error.AppError(f"{self.name} is not ready")
        self.handler(*args, **kwargs)  # pylint: disable=not-callable
        self.handler.beforeInitHooks.extend(self.beforeInitHooks)
        self.handler.afterInitHooks.extend(self.afterInitHooks)
        self.handler.beforeCallHooks.extend(self.beforeCallHooks)
        self.handler.afterSaveHooks.extend(self.afterCallHooks)
        self.handler.beforeExitHooks.extend(self.beforeExitHooks)
        return self

    def start(self, *args, **kwargs):
        if not self.handler:
            raise error.AppError(f"{self.name} is not ready")
        self.handler.start(*args, **kwargs)
        return self

    def input(self, argument):
        def _dec(funcOrApp):
            funcOrApp = self.handler if isinstance(funcOrApp, SparkApp) else funcOrApp
            self.handler = Handler.input(argument)(funcOrApp)
            return self

        return _dec

    def output(self, argument):
        def _dec(funcOrApp):
            funcOrApp = self.handler if isinstance(funcOrApp, SparkApp) else funcOrApp
            self.handler = Handler.output(argument)(funcOrApp)
            return self

        return _dec

    def param(self, argument):
        def _dec(funcOrApp):
            funcOrApp = self.handler if isinstance(funcOrApp, SparkApp) else funcOrApp
            self.handler = Handler.param(argument)(funcOrApp)
            return self

        return _dec

    def column(self, argument):
        def _dec(funcOrApp):
            funcOrApp = self.handler if isinstance(funcOrApp, SparkApp) else funcOrApp
            self.handler = Handler.column(argument)(funcOrApp)
            return self

        return _dec



# coding=utf-8
from __future__ import absolute_import, print_function

import itertools

from suanpan.arguments import Bool, Float, Int, String
from suanpan.mq import redis
from suanpan.proxy import Proxy


class MQProxy(Proxy):
    MAPPING = {"redis": redis.MQ}
    DEFAULT_ARGUMENTS = [String("mq-type")]
    REDIS_ARGUMENTS = [
        String("mq-redis-host", default="localhost"),
        Int("mq-redis-port", default=6379),
        Bool("mq-redis-realtime", default=False),
        Bool("mq-redis-keepalive", default=True),
        # Int("mq-redis-keepalive-idle", default=120),
        # Int("mq-redis-keepalive-cnt", default=2),
        # Int("mq-redis-keepalive-intvl", default=30),
        Float("mq-redis-socket-connect-timeout", default=1),
        String("mq-redis-unix-socket-path"),
    ]
    ARGUMENTS = list(itertools.chain(DEFAULT_ARGUMENTS, REDIS_ARGUMENTS))


mq = MQProxy()



# coding=utf-8
from __future__ import absolute_import, print_function

import abc
import functools
import os
import tempfile
import zipfile

from suanpan import path as spath
from suanpan.log import logger
from suanpan.objects import HasName


class Storage(HasName):
    __metaclass__ = abc.ABCMeta

    DEFAULT_IGNORE_KEYWORDS = ("__MACOSX", ".DS_Store")
    CONTENT_MD5 = "Content-MD5"
    PBAR_FORMAT = "{l_bar}{bar}"

    def __init__(
        self, delimiter=os.sep, tempStore=tempfile.gettempdir(), **kwargs
    ):  # pylint: disable=unused-argument
        self.delimiter = delimiter
        self.tempStore = tempStore

    @abc.abstractmethod
    def download(self, name, path, **kwargs):
        pass

    @abc.abstractmethod
    def upload(self, name, path, **kwargs):
        pass

    @abc.abstractmethod
    def copy(self, name, dist, **kwargs):
        pass

    @abc.abstractmethod
    def remove(self, name, **kwargs):
        pass

    @abc.abstractmethod
    def walk(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listAll(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listFolders(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listFiles(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def isFolder(self, path, **kwargs):
        pass

    @abc.abstractmethod
    def isFile(self, path, **kwargs):
        pass

    def compress(self, zipFilePath, path, ignores=DEFAULT_IGNORE_KEYWORDS):
        compressFunc = self.compressFolder if os.path.isdir(path) else self.compressFile
        return compressFunc(zipFilePath, path, ignores=ignores)

    def compressFolder(self, zipFilePath, folderPath, ignores=DEFAULT_IGNORE_KEYWORDS):
        if folderPath in ignores:
            logger.info(f"Ignore compressing folder: {folderPath} -> {zipFilePath}")
            return zipFilePath

        logger.info(f"Compressing folder: {folderPath} -> {zipFilePath}")
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            for root, _, files in os.walk(folderPath):
                for file in files:
                    filePath = os.path.join(root, file)
                    zip.write(
                        filePath, arcname=self.localRelativePath(filePath, folderPath)
                    )
        logger.info(f"Compressed folder: {folderPath} -> {zipFilePath}")
        return zipFilePath

    def compressFile(self, zipFilePath, filePath, ignores=DEFAULT_IGNORE_KEYWORDS):
        if filePath in ignores:
            logger.info(f"Ignore compressing File: {filePath} -> {zipFilePath}")
            return zipFilePath

        logger.info(f"Compressing File: {filePath} -> {zipFilePath}")
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            _, filename = os.path.split(filePath)
            zip.write(filePath, arcname=filename)
        logger.info(f"Compressed File: {filePath} -> {zipFilePath}")
        return zipFilePath

    def extract(self, zipFilePath, distPath, ignores=DEFAULT_IGNORE_KEYWORDS):
        logger.info(f"Extracting zip: {zipFilePath} -> {distPath}")
        with zipfile.ZipFile(zipFilePath, "r") as zip:
            zip.extractall(distPath)
        self.removeIgnores(distPath, ignores=ignores)
        logger.info(f"Extracted zip: {zipFilePath} -> {distPath}")

    def getPathInTempStore(self, path, tempStore=None):
        tempStore = tempStore or self.tempStore
        return self.localPathJoin(tempStore, path)

    def completePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path if path.endswith(delimiter) else path + delimiter

    def toLocalPath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return str(path).replace(delimiter, os.sep)

    def toStoragePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return str(path).replace(os.sep, delimiter)

    def localPathJoin(self, *paths):
        path = os.path.join(*paths)
        return self.toLocalPath(path)

    def storagePathJoin(self, *paths):
        path = os.path.join(*[self.toLocalPath(path) for path in paths])
        return self.toStoragePath(path)

    def localRelativePath(self, path, base):
        return self._relativePath(path, base, delimiter=os.sep)

    def storageRelativePath(self, path, base, delimiter=None):
        delimiter = delimiter or self.delimiter
        return self._relativePath(path, base, delimiter=delimiter)

    def _relativePath(self, path, base, delimiter):
        base = base if base.endswith(delimiter) else base + delimiter
        return path[len(base) :] if path.startswith(base) else path

    @abc.abstractmethod
    def getStorageMd5(self, name, **kwargs):
        pass

    def getLocalMd5(self, path, **kwargs):  # pylint: disable=unused-argument
        return spath.md5(path) if os.path.isfile(path) else None

    def checkMd5(self, md5a, md5b, **kwargs):  # pylint: disable=unused-argument
        return md5a if md5a == md5b and md5a is not None else False

    @abc.abstractmethod
    def getStorageSize(self, name, **kwargs):  # pylint: disable=unused-argument
        pass

    def getLocalSize(self, path, **kwargs):  # pylint: disable=unused-argument
        return os.path.getsize(path)

    def storageUrl(self, path, **kwargs):  # pylint: disable=unused-argument
        return "storage://" + path

    def removeIgnores(self, path, ignores=None):
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS

        def _ignore(_path):
            spath.remove(_path)
            logger.info(f"Removed ignored: {_path}")
            return _path

        def _getIgnores(path):
            for root, folders, files in os.walk(path):
                for folder in folders:
                    if folder in ignores:
                        yield os.path.join(root, folder)
                for file in files:
                    if file in ignores:
                        yield os.path.join(root, file)

        return [_ignore(_path) for _path in _getIgnores(path)]

    def pathSplit(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path.rsplit(delimiter, 1)

    def pathSplitExt(self, path, extDelimiter="."):
        return path.rsplit(extDelimiter, 1)

    def pbarRunner(self, pbar, quantity=1):
        def _dec(runner):
            @functools.wraps(runner)
            def _runner(*args, **kwargs):
                result = runner(*args, **kwargs)
                pbar.update(quantity)
                return result

            return _runner

        return _dec



# coding=utf-8
from __future__ import absolute_import, print_function

import functools
import sys
import traceback

import retrying
import timeout_decorator

from suanpan.log import logger


def needRetryException(exception):
    return not isinstance(exception, (KeyboardInterrupt, SystemExit))


def retryRunner(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.warning(f"Run failed and retrying: {func.__name__}")
            logger.warning(traceback.format_exc())
            raise e

    return _dec


def retry(*args, **kwargs):
    def _wrap(func):
        kwargs.update(wrap_exception=True, retry_on_exception=needRetryException)
        _retry = retrying.retry(*args, **kwargs)
        _func = _retry(retryRunner(func))

        @functools.wraps(func)
        def _dec(*fargs, **fkwargs):
            try:
                return _func(*fargs, **fkwargs)
            except retrying.RetryError as e:
                _, error, _ = e.last_attempt.value
                if needRetryException(error):
                    rfunc = func.func if isinstance(func, functools.partial) else func
                    rfuncname = getattr(rfunc, "__name__", "unknown_func")
                    logger.error(
                        f"Retry failed after {e.last_attempt.attempt_number} attempts: {rfuncname}"
                    )
                    raise e
                raise error

        return _dec

    return _wrap


def globalrun(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            logger.debug("User canceled and exit 0.")
            sys.exit(0)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(traceback.format_exc())
            raise e

    return _dec


def saferun(func, default=None):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            logger.debug("User canceled and exit 0.")
            sys.exit(0)
        except Exception:  # pylint: disable=broad-except
            logger.warning(f"Ignore Error: \n{traceback.format_exc()}")
            return default

    return _dec


timeout = timeout_decorator.timeout



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import copy
import itertools
import re
from collections import namedtuple

from lostc import collection as lcc

from suanpan import error, g
from suanpan.log import logger
from suanpan.objects import HasName
from suanpan.utils import json

Param = namedtuple("Param", ["uuid", "name", "type"])
Port = namedtuple("Port", ["uuid", "name", "type", "subtype"])


class BaseNode(HasName):
    NODE_INFO_KEY = "SP_NODE_INFO"
    DEFAULT_NODE_INFO = {"info": {}, "inputs": {}, "outputs": {}, "params": {}}

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(BaseNode, self).__init__()
        self._node = self.load()

    def __getattr__(self, key):
        value = self._getPort(key) if re.match(r"(in|out)\d+", key) else self.get(key)
        if value is None:
            raise error.NodeError(f"{self.name} not contains {key}")
        return value

    def _getPort(self, key):
        return lcc.first(
            itertools.chain(self.inputs, self.outputs), lambda i: i.uuid == key
        )

    def get(self, key, default=None):
        for _, collection in self._node.items():
            if isinstance(collection, dict):
                if key in collection:
                    return collection[key]
        return default

    @property
    def info(self):
        return self._node["info"]

    @property
    def inputs(self):
        return self._node["inputs"].values()

    @property
    def ins(self):
        return self._node["inputs"].values()

    @property
    def outputs(self):
        return self._node["outputs"].values()

    @property
    def outs(self):
        return self._node["outputs"].values()

    @property
    def params(self):
        return self._node["params"].values()

    def loadFromEnv(self):
        nodeInfoBase64 = g.nodeInfo
        logger.debug(f"{self.NODE_INFO_KEY}(Base64)='{nodeInfoBase64}'")
        nodeInfoString = base64.b64decode(nodeInfoBase64).decode()
        nodeInfo = json.loads(nodeInfoString)
        return nodeInfo

    def formatNodeInfo(self, nodeInfo):
        inputs = {
            name: Port(**port) for name, port in nodeInfo.get("inputs", {}).items()
        }
        outputs = {
            name: Port(**port) for name, port in nodeInfo.get("outputs", {}).items()
        }
        params = {
            name: Param(**param) for name, param in nodeInfo.get("params", {}).items()
        }
        return {"inputs": inputs, "outputs": outputs, "params": params}

    def defaultNodeInfo(self):
        return copy.deepcopy(self.DEFAULT_NODE_INFO)

    def _updateInfo(self, *infos):
        result = self.defaultNodeInfo()
        keys = result.keys()
        for info in infos:
            for key in keys:
                result[key].update(info.get(key, {}))
        return result

    def load(self):
        return self._updateInfo(self.formatNodeInfo(self.loadFromEnv()))



# coding=utf-8
from __future__ import absolute_import, print_function

import itertools

from suanpan.arguments import Bool, Float, Int, String
from suanpan.mstorage import redis
from suanpan.proxy import Proxy


class MStorageProxy(Proxy):
    MAPPING = {"redis": redis.MStorage}
    DEFAULT_ARGUMENTS = [String("mstorage-type")]
    REDIS_ARGUMENTS = [
        String("mstorage-redis-host", default="localhost"),
        Int("mstorage-redis-port", default=6379),
        Bool("mstorage-redis-keepalive", default=True),
        Int("mstorage-redis-keepalive-idle", default=120),
        Int("mstorage-redis-keepalive-cnt", default=2),
        Int("mstorage-redis-keepalive-intvl", default=30),
        Int("mstorage-redis-expire", default=30),
        Float("mstorage-redis-socket-connect-timeout", default=1),
        String("mstorage-redis-unix-socket-path"),
    ]
    ARGUMENTS = list(itertools.chain(DEFAULT_ARGUMENTS, REDIS_ARGUMENTS))


mstorage = MStorageProxy()



# coding=utf-8
from __future__ import absolute_import, print_function

import redis

from suanpan import error


class MStorage(object):
    def __init__(
        self,  # pylint: disable=unused-argument
        redisHost="localhost",
        redisPort=6379,
        redisKeepalive=False,
        # redisKeepaliveIDLE=120,
        # redisKeepaliveCNT=2,
        # redisKeepaliveINTVL=30,
        redisDefaultExpire=30,
        redisSocketConnectTimeout=1,
        redisUnixSocketPath=None,
        options=None,
        client=None,
        **kwargs
    ):
        self.options = options or {}
        self.options.update(
            host=redisHost,
            port=redisPort,
            keepalive=redisKeepalive,
            # keepidle=redisKeepaliveIDLE,
            # keepcnt=redisKeepaliveCNT,
            # keepintvl=redisKeepaliveINTVL,
            socketConnectTimeout=redisSocketConnectTimeout,
            unixSocketPath=redisUnixSocketPath,
        )
        if client and not isinstance(client, redis.Redis):
            raise error.MStorageError(f"Invalid client {client}")

        if client:
            self.client = client
        elif self.options["unixSocketPath"]:
            self.client = redis.Redis(
                decode_responses=True, unix_socket_path=self.options["unixSocketPath"]
            )
        else:
            self.client = redis.Redis(
                host=self.options["host"],
                port=self.options["port"],
                decode_responses=True,
                socket_keepalive=self.options["keepalive"],
                # socket_keepalive_options={
                #     socket.TCP_KEEPIDLE: self.options["keepidle"],
                #     socket.TCP_KEEPCNT: self.options["keepcnt"],
                #     socket.TCP_KEEPINTVL: self.options["keepintvl"],
                # },
                socket_connect_timeout=self.options["socketConnectTimeout"],
            )

        self.defaultExpire = redisDefaultExpire

    def type(self, name):
        return self.client.type(name)

    def exists(self, *names):
        return self.client.exists(*names)

    def expire(self, name, expire):
        return self.client.expire(name, expire)

    def delete(self, *names):
        return self.client.delete(*names)

    def get(self, name):
        return self._kget(name)

    def set(self, name, value, *args, **kwargs):
        kwargs.setdefault("expire", self.defaultExpire)
        return self._kset(name, value, *args, **kwargs)

    def mget(self, name):
        return self._hmgetall(name)

    def mset(self, name, mapping, expire=None):
        result = self._hmset(name, mapping)
        expire = expire or self.defaultExpire
        self.expire(name, expire)
        return result

    def lpush(self, name, *values, **kwargs):
        first = kwargs.pop("first", False)
        _push = self._lpush if first else self._rpush
        return _push(name, *values)

    def lpop(self, name, count=1, first=True):
        return (
            self._lpopmore(name, count, first=first)
            if count > 1
            else self._lpopone(name, first=first)
        )

    def llen(self, name):
        return self._llen(name)

    def lrange(self, name, start, end):
        return self._lrange(name, start, end)

    def ltrim(self, name, start, end):
        return self._ltrim(name, start, end)

    def _lpopone(self, name, first=True):
        _pop = self._lpop if first else self._rpop
        return _pop(name)

    def _lpopmore(self, name, count, first=True):
        length = self._llen(name)
        if not length:
            return []

        count = min(count, length)
        poprange, trimrange = (
            ((0, count - 1), (count, -1))
            if first
            else ((-count, -1), (0, length - count - 1))
        )
        values = self._lrange(name, *poprange)
        self._ltrim(name, *trimrange)
        return values

    def _kget(self, name):
        return self.client.get(name)

    def _kset(self, name, value, expire=None):
        return self.client.set(name, value, ex=expire)

    def _hmget(self, name, keys=None):
        return self._hmgetfields(name, keys) if keys else self._hmgetall(name)

    def _hmgetall(self, name):
        return self._decode_keys(self.client.hgetall(name))

    def _hmgetfields(self, name, keys):
        return self._decode_keys(dict(zip(keys, self.client.hmget(name, keys))))

    def _hmset(self, name, mapping):
        return self.client.hmset(name, mapping)

    def _lset(self, name, index, value):
        return self.client.lset(name, index, value)

    def _llen(self, name):
        return self.client.llen(name)

    def _lpop(self, name):
        return self.client.lpop(name)

    def _rpop(self, name):
        return self.client.rpop(name)

    def _lpush(self, name, *values):
        return self.client.lpush(name, *values)

    def _rpush(self, name, *values):
        return self.client.rpush(name, *values)

    def _lrange(self, name, start, end):
        return self.client.lrange(name, start, end)

    def _ltrim(self, name, start, end):
        return self.client.ltrim(name, start, end)

    def _decode_keys(self, mapping):
        return {k.decode(): v for k, v in mapping.items()}



# coding=utf-8
from __future__ import absolute_import, print_function

import functools
import itertools
import time

from suanpan import utils
from suanpan.log import logger


def formatFuncCall(func, *args, **kwargs):
    paramString = ", ".join(
        itertools.chain((str(a) for a in args), (f"{k}={v}" for k, v in kwargs.items()))
    )
    funcString = f"{func.__name__}({utils.shorten(paramString)})"
    return funcString


def costCall(func, *args, **kwargs):
    startTime = time.time()
    result = func(*args, **kwargs)
    endTime = time.time()
    costTime = endTime - startTime
    return costTime, result


def cost(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        costTime, result = costCall(func, *args, **kwargs)
        logger.debug(f"{formatFuncCall(func, *args, **kwargs)} - {costTime}s")
        return result

    return _dec


def print(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        result = func(*args, **kwargs)
        logger.debug(f"{formatFuncCall(func, *args, **kwargs)} - {result}")
        return result

    return _dec
